use crate::{DBError, DBResult};

use std::{
    collections::HashMap,
    sync::{Arc, Mutex},
};

pub type DBConnection = std::sync::Arc<Connection>;

pub struct Connection {
    // we leak the ConnectionThreadSafe and make sure that the only references to it are stored in
    // statement_cache, so as long as we drop the statement_cache first there are no correctness
    // issues, and we get to have static-lifetime statements for caching purposes.
    conn: &'static sqlite::ConnectionThreadSafe,
    statement_cache: Mutex<HashMap<u64, sqlite::Statement<'static>>>,
}

impl Connection {
    pub fn open<U: AsRef<str>>(uri: U) -> Result<DBConnection, DBError> {
        match sqlite::Connection::open_thread_safe_with_flags(
            uri.as_ref(),
            sqlite::OpenFlags::new()
                .with_create()
                .with_full_mutex()
                .with_read_write(),
        ) {
            Ok(conn) => {
                let conn = Box::leak(Box::new(conn));
                Ok(Arc::new(Self {
                    conn,
                    statement_cache: Default::default(),
                }))
            }
            Err(e) => Err(DBError::Sqlite(e)),
        }
    }

    pub(crate) fn execute_raw(&self, sql: &str) -> Result<(), DBError> {
        Ok(self.conn.execute(sql)?)
    }

    pub(crate) fn with_prepared<R>(
        &self,
        hash_key: u64,
        build_query: impl Fn() -> String,
        run_query: impl Fn(&mut sqlite::Statement<'static>) -> DBResult<R>,
    ) -> DBResult<R> {
        match self.statement_cache.lock()?.entry(hash_key) {
            std::collections::hash_map::Entry::Vacant(e) => {
                let q: sqlite::Statement<'static> = self.conn.prepare(build_query())?;
                run_query(e.insert(q))
            }
            std::collections::hash_map::Entry::Occupied(mut e) => run_query(e.get_mut()),
        }
    }
}

impl Drop for Connection {
    fn drop(&mut self) {
        // first we drop the statement cache, if we can get it
        // if this fails then we're entering UD, but we probably panicked anyway
        self.statement_cache.lock().map(|mut sc| sc.clear()).ok();
        // then we drop the connection
        unsafe {
            drop(Box::from_raw(
                self.conn as *const _ as *mut sqlite::ConnectionThreadSafe,
            ));
        }
    }
}
